@@ -34,6 +34,15 @@ def _hadd(root_files, dst, dry=False):
     """
     Compiles and runs the hadd command
     """
+    if len(root_files) == 1:
+        # No need for hadding, just copy
+        rootfile = root_files[0]
+        logger.debug('Just 1 rootfile - copying %s --> %s', rootfile, dst)
+        if seutils.has_protocol(rootfile):
+            seutils.cp(rootfile, dst)
+        else:
+            shutil.copyfile(rootfile, dst)
+        return
     cmd = ['hadd', '-f', dst] + root_files
     if dry:
         logger.warning('hadd command: ' + ' '.join(cmd))
@@ -70,14 +79,11 @@ def _hadd_chunks(root_files, dst, n_threads=6, chunk_size=200, tmpdir='/tmp', dr
         # No need for chunking. This should also be the final step of the recursion
         _hadd(root_files, dst, dry=dry)
         return
-
     import multiprocessing as mp
     n_chunks = int(math.ceil(len(root_files) / float(chunk_size)))
-
     # Make a unique directory for temporary files
     tmpdir = osp.join(tmpdir, 'tmphadd', str(uuid.uuid4()))
     os.makedirs(tmpdir)
-
     try:
         debug(True)
         chunk_rootfiles = []
@@ -91,13 +97,13 @@ def _hadd_chunks(root_files, dst, n_threads=6, chunk_size=200, tmpdir='/tmp', dr
             if dry: logger.debug('hadding %s --> %s', ' '.join(chunk), chunk_dst)
         # Submit to multiprocessing in one go:
         if not dry:
+            n_threads = min(n_threads, len(n_chunks)) # Don't start more threads if not necessary
             p = mp.Pool(n_threads)
             p.map(_hadd_packed, func_args)
             p.close()
             p.join()
         # Merge the chunks into the final destination, potentially with another chunked merge
         _hadd_chunks(chunk_rootfiles, dst, n_threads, chunk_size, tmpdir, dry)
-
     finally:
         logger.warning('Removing %s', tmpdir)
         shutil.rmtree(tmpdir)
@@ -107,24 +113,32 @@ def _hadd_chunks(root_files, dst, n_threads=6, chunk_size=200, tmpdir='/tmp', dr
 # Root utilities
 
 USE_CACHE=False
+CACHE_DIR=None
 CACHE_NENTRIES=None
 CACHE_TREESINFILE=None
 
-def use_cache(flag=True):
+def use_cache(flag=True, cache_dir=None):
+    global CACHE_DIR
     global CACHE_NENTRIES
     global CACHE_TREESINFILE
     global USE_CACHE
     USE_CACHE = flag
+    CACHE_DIR = cache_dir
     if flag:
         from .cache import FileCache
-        if CACHE_NENTRIES is None:        
-            CACHE_NENTRIES = FileCache('seutils.nentries')
+        if CACHE_NENTRIES is None:
+            CACHE_NENTRIES = FileCache('seutils.nentries', app_cache_dir=cache_dir)
+            # Do save whatever the chosen default was
+            if CACHE_DIR is None: CACHE_DIR = CACHE_NENTRIES.app_cache_dir
         if CACHE_TREESINFILE is None:        
-            CACHE_TREESINFILE = FileCache('seutils.treesinfile')
+            CACHE_TREESINFILE = FileCache('seutils.treesinfile', app_cache_dir=cache_dir)
     else:
         USE_CACHE=False
         CACHE_NENTRIES=None
         CACHE_TREESINFILE=None
+        CACHE_DIR=None
+    logger.info('Activated cache using directory %s', CACHE_DIR)
+    return CACHE_DIR
 
 @contextmanager
 def nocache():
@@ -132,20 +146,78 @@ def nocache():
     Context manager to temporarily disable the cache
     """
     global USE_CACHE
+    global CACHE_DIR
     global CACHE_NENTRIES
     global CACHE_TREESINFILE
+    _saved_CACHE_DIR = CACHE_DIR
     _saved_CACHE_NENTRIES = CACHE_NENTRIES
     _saved_CACHE_TREESINFILE = CACHE_TREESINFILE
     USE_CACHE = False
+    CACHE_DIR = None
     CACHE_NENTRIES = None
     CACHE_TREESINFILE = None
     try:
         yield None
     finally:
         USE_CACHE = True
+        CACHE_DIR = _saved_CACHE_DIR
         CACHE_NENTRIES = _saved_CACHE_NENTRIES
         CACHE_TREESINFILE = _saved_CACHE_TREESINFILE
 
+@contextmanager
+def cache(cache_dir=None):
+    """
+    Context manager to temporarily enable the cache in some directory.
+    If cache_dir is None, some unique cache_dir is used
+    """
+    global USE_CACHE
+    global CACHE_DIR
+    global CACHE_NENTRIES
+    global CACHE_TREESINFILE
+    _saved_USE_CACHE = USE_CACHE
+    _saved_CACHE_DIR = CACHE_DIR
+    _saved_CACHE_NENTRIES = CACHE_NENTRIES
+    _saved_CACHE_TREESINFILE = CACHE_TREESINFILE
+    use_cache(cache_dir=cache_dir)
+    try:
+        yield cache_dir
+    finally:
+        USE_CACHE = _saved_USE_CACHE
+        CACHE_DIR = _saved_CACHE_DIR
+        CACHE_NENTRIES = _saved_CACHE_NENTRIES
+        CACHE_TREESINFILE = _saved_CACHE_TREESINFILE
+
+def cache_to_file(dst='rootcache.tar.gz'):
+    """
+    Dumps the cached requests to a file
+    """
+    if not USE_CACHE:
+        raise RuntimeError('No active cache to save to a file')
+    if not dst.endswith('.tar.gz'): dst += '.tar.gz'
+    dst = osp.abspath(dst)
+    try:
+        _return_dir = os.getcwd()
+        os.chdir(CACHE_DIR)
+        cmd = ['tar', '-zcvf', dst, '.']
+        seutils.run_command(cmd)
+        return dst
+    finally:
+        os.chdir(_return_dir)
+
+def load_cache(cache_file, dst='.rootcache'):
+    """
+    Loads to use the cache from a .tar.gz file
+    """
+    dst = osp.abspath(dst)
+    logger.info('Extracting %s --> %s', cache_file, dst)
+    if not osp.isdir(dst): os.makedirs(dst)
+    cmd = [
+        'tar', '-xvf', cache_file,
+        '-C', dst
+        ]
+    seutils.run_command(cmd)
+    use_cache(cache_dir=dst)
+
 @contextmanager
 def open_root(rootfile, mode='read'):
     """
@@ -199,6 +271,7 @@ def _get_trees_cache(rootfile):
     global USE_CACHE
     # Check if cache should be used and whether it exists
     if not USE_CACHE:
+        logger.debug('Not using cached trees: USE_CACHE==False')
         return None
     # If rootfile is TDirectory-like, use that
     try:
@@ -207,8 +280,9 @@ def _get_trees_cache(rootfile):
         pass
     if rootfile in CACHE_TREESINFILE:
         trees = CACHE_TREESINFILE[rootfile]
-        logger.info('Using cached result trees in %s: %s', rootfile, trees)
+        logger.info('Using cached trees in %s: %s', rootfile, trees)
         return trees
+    logger.debug('Not using cached trees: No cached result for %s', rootfile)
     return None        
 
 def _select_most_likely_tree(trees):
@@ -243,6 +317,7 @@ def get_trees(node):
     nodename = node if is_string(node) else node.GetPath()
     cached_result = _get_trees_cache(nodename)
     if cached_result:
+        logger.debug('Using cached trees %s', cached_result)
         return cached_result
     logger.debug('No cached result for %s', nodename)
     if is_string(node):
@@ -252,6 +327,7 @@ def get_trees(node):
         trees = _get_trees_recursively_root(node)
     # Update the cache
     if USE_CACHE:
+        logger.debug('Caching trees for %s: %s', nodename, trees)
         CACHE_TREESINFILE[nodename] = trees
         CACHE_TREESINFILE.sync()
     return trees
@@ -263,13 +339,7 @@ def get_most_likely_tree(node):
     trees = get_trees(node)
     return _select_most_likely_tree(trees)
 
-def _count_entries_root(tfile, tree='auto'):
-    if tree == 'auto':
-        trees = _get_trees_recursively_root(tfile)
-        if len(trees) == 0:
-            logger.error('No TTrees found in %s', tfile)
-            return None
-        tree = _select_most_likely_tree(trees)
+def _count_entries_root(tfile, tree):
     ttree = tfile.Get(tree)
     return ttree.GetEntries()
 
@@ -297,6 +367,12 @@ def count_entries(rootfile, tree='auto'):
         if nentries: return nentries
     # At least some part of the cached couldn't return, so open the root file
     with open_root(rootfile) as tfile:
+        if tree == 'auto':
+            trees = get_trees(rootfile)
+            if len(trees) == 0:
+                logger.error('No TTrees found in %s', tfile)
+                return None
+            tree = _select_most_likely_tree(trees)
         nentries = _count_entries_root(tfile, tree)
     # Cache the result
     if USE_CACHE:
@@ -386,7 +462,7 @@ def _get_chunkify_nentries(nentries, n_chunks):
 
 def iter_chunkify_rootfiles_by_entries(rootfiles, chunk_size):
     """
-    Function that yields lists of (path.root, n_take, first, last), so that the number
+    Function that yields lists of (rootfile, first, last, is_whole_rootfile), so that the number
     of events in a list is equal to chunk_size
     """
     if not len(rootfiles): raise StopIteration
@@ -401,11 +477,12 @@ def iter_chunkify_rootfiles_by_entries(rootfiles, chunk_size):
     while True:
         n_take = min(n_still_needed, nentries)
         nentries -= n_take
+        is_whole_rootfile = (first == 0 and nentries == 0)
         logger.debug(
             'Taking %s entries from %s (%s - %s); %s remaining',
             n_take, rootfile, first, first+n_take-1, nentries
             )
-        chunk.append((rootfile, first, first+n_take-1))
+        chunk.append((rootfile, first, first+n_take-1, is_whole_rootfile))
         n_still_needed -= n_take
         first += n_take
         if n_still_needed == 0:
@@ -429,19 +506,55 @@ def hadd_chunk_entries(chunk, dst, file_split_fn=make_chunk_rootfile, tree='auto
     (list of (rootfile, first, last)), and merges it into
     one rootfile at path dst
     """
-    tmpdir = 'merging-{0}'.format(uuid.uuid4())
-    os.makedirs(tmpdir)
-    try:
-        # First created the splitted rootfiles
-        splitted_rootfiles = []
-        for i, (rootfile, first, last) in enumerate(chunk):
-            splitted_rootfile = osp.join(tmpdir, 'part{}.root'.format(i))
+    if len(chunk) == 1:
+        # No need for the hadding step, just run the split directly to dst
+        rootfile, first, last, is_whole_rootfile = chunk[0]
+        if is_whole_rootfile:
+            logger.info(
+                'Detected chunk %s is one *whole* file; Simply copying to %s directly',
+                chunk, dst
+                )
+            seutils.cp(rootfile, dst)
+        else:
+            logger.info(
+                'Detected chunk %s is one file; Splitting (%s-%s) to %s directly',
+                chunk, first, last, dst
+                )
             file_split_fn(
                 rootfile, first, last,
-                dst=splitted_rootfile, tree=tree
-                )
-            splitted_rootfiles.append(splitted_rootfile)
+                dst=dst, tree=tree
+                )            
+    try:
+        tmpdir = osp.abspath('merging-{0}'.format(uuid.uuid4()))
+        os.makedirs(tmpdir)
+        # First created the splitted rootfiles
+        splitted_rootfiles = []
+        for i, (rootfile, first, last, is_whole_rootfile) in enumerate(chunk):
+            if is_whole_rootfile:
+                # No need for splitting, just use the whole root file
+                if seutils.has_protocol(rootfile):
+                    splitted_rootfile = osp.join(tmpdir, 'part{}.root'.format(i))
+                    logger.debug(
+                        'rootfile %s is used entirely - copying to %s',
+                        rootfile, splitted_rootfile
+                        )
+                    seutils.cp(rootfile, splitted_rootfile)
+                    splitted_rootfiles.append(splitted_rootfile)
+                else:
+                    # If the file is local, don't even copy it
+                    logger.debug('local rootfile %s is used entirely - using it as is', rootfile)
+                    splitted_rootfiles.append(rootfile)
+            else:
+                splitted_rootfile = osp.join(tmpdir, 'part{}.root'.format(i))
+                file_split_fn(
+                    rootfile, first, last,
+                    dst=splitted_rootfile, tree=tree
+                    )
+                splitted_rootfiles.append(splitted_rootfile)
         # Hadd the splitted rootfiles
         hadd_chunks(splitted_rootfiles, dst, tmpdir=tmpdir)
     finally:
-        shutil.rmtree(tmpdir)
+        try:
+            shutil.rmtree(tmpdir)
+        except:
+            pass
